@@ -95,11 +95,14 @@ class _BlockedThread(object):
 
     __bool__ = __nonzero__
 
-    def block(self):
+    def block(self, deadline=None):
         if self._blocked_thread is not None:
             raise Exception("Cannot block more than one thread on one BlockedThread")
         self._blocked_thread = greenlet.getcurrent()
 
+        if deadline is not None:
+            self._hub.schedule_call_local(deadline - self._hub.clock(), self.wake)
+
         try:
             self._hub.switch()
         finally:
@@ -245,6 +248,7 @@ class Socket(_Socket):
                                                       event,
                                                       lambda _: None,
                                                       lambda: None)
+        self.__dict__['_eventlet_clock'] = hub.clock
 
     @_wraps(_Socket.close)
     def close(self, linger=None):
@@ -320,46 +324,46 @@ class Socket(_Socket):
             return _Socket_send_multipart(self, msg_parts, flags, copy, track)
 
     @_wraps(_Socket.send_string)
-    def send_string(self, msg_parts, flags=0, copy=True, track=False):
+    def send_string(self, u, flags=0, copy=True, encoding='utf-8'):
         """A send_string method that's safe to use when multiple
         greenthreads are calling send, send_string, recv and
         recv_string on the same socket.
         """
         if flags & NOBLOCK:
-            return _Socket_send_string(self, msg_parts, flags, copy, track)
+            return _Socket_send_string(self, u, flags, copy, encoding)
 
         # acquire lock here so the subsequent calls to send for the
         # message parts after the first don't block
         with self._eventlet_send_lock:
-            return _Socket_send_string(self, msg_parts, flags, copy, track)
+            return _Socket_send_string(self, u, flags, copy, encoding)
 
     @_wraps(_Socket.send_pyobj)
-    def send_pyobj(self, msg_parts, flags=0, copy=True, track=False):
+    def send_pyobj(self, obj, flags=0, protocol=2):
         """A send_pyobj method that's safe to use when multiple
         greenthreads are calling send, send_pyobj, recv and
         recv_pyobj on the same socket.
         """
         if flags & NOBLOCK:
-            return _Socket_send_pyobj(self, msg_parts, flags, copy, track)
+            return _Socket_send_pyobj(self, obj, flags, protocol)
 
         # acquire lock here so the subsequent calls to send for the
         # message parts after the first don't block
         with self._eventlet_send_lock:
-            return _Socket_send_pyobj(self, msg_parts, flags, copy, track)
+            return _Socket_send_pyobj(self, obj, flags, protocol)
 
     @_wraps(_Socket.send_json)
-    def send_json(self, msg_parts, flags=0, copy=True, track=False):
+    def send_json(self, obj, flags=0, **kwargs):
         """A send_json method that's safe to use when multiple
         greenthreads are calling send, send_json, recv and
         recv_json on the same socket.
         """
         if flags & NOBLOCK:
-            return _Socket_send_json(self, msg_parts, flags, copy, track)
+            return _Socket_send_json(self, obj, flags, **kwargs)
 
         # acquire lock here so the subsequent calls to send for the
         # message parts after the first don't block
         with self._eventlet_send_lock:
-            return _Socket_send_json(self, msg_parts, flags, copy, track)
+            return _Socket_send_json(self, obj, flags, **kwargs)
 
     @_wraps(_Socket.recv)
     def recv(self, flags=0, copy=True, track=False):
@@ -376,6 +380,16 @@ class Socket(_Socket):
             self._eventlet_recv_event.wake()
             return msg
 
+        deadline = None
+        if hasattr(__zmq__, 'RCVTIMEO'):
+            sock_timeout = self.getsockopt(__zmq__.RCVTIMEO)
+            if sock_timeout == -1:
+                pass
+            elif sock_timeout > 0:
+                deadline = self._eventlet_clock() + sock_timeout / 1000.0
+            else:
+                raise ValueError(sock_timeout)
+
         flags |= NOBLOCK
         with self._eventlet_recv_lock:
             while True:
@@ -383,7 +397,12 @@ class Socket(_Socket):
                     return _Socket_recv(self, flags, copy, track)
                 except ZMQError as e:
                     if e.errno == EAGAIN:
-                        self._eventlet_recv_event.block()
+                        # zmq in its wisdom decided to reuse EAGAIN for timeouts
+                        if deadline is not None and self._eventlet_clock() > deadline:
+                            e.is_timeout = True
+                            raise
+
+                        self._eventlet_recv_event.block(deadline=deadline)
                     else:
                         raise
                 finally:
@@ -407,43 +426,43 @@ class Socket(_Socket):
             return _Socket_recv_multipart(self, flags, copy, track)
 
     @_wraps(_Socket.recv_string)
-    def recv_string(self, flags=0, copy=True, track=False):
+    def recv_string(self, flags=0, encoding='utf-8'):
         """A recv_string method that's safe to use when multiple
         greenthreads are calling send, send_string, recv and
         recv_string on the same socket.
         """
         if flags & NOBLOCK:
-            return _Socket_recv_string(self, flags, copy, track)
+            return _Socket_recv_string(self, flags, encoding)
 
         # acquire lock here so the subsequent calls to recv for the
         # message parts after the first don't block
         with self._eventlet_recv_lock:
-            return _Socket_recv_string(self, flags, copy, track)
-
-    @_wraps(_Socket.recv_pyobj)
-    def recv_pyobj(self, flags=0, copy=True, track=False):
-        """A recv_pyobj method that's safe to use when multiple
-        greenthreads are calling send, send_pyobj, recv and
-        recv_pyobj on the same socket.
-        """
-        if flags & NOBLOCK:
-            return _Socket_recv_pyobj(self, flags, copy, track)
-
-        # acquire lock here so the subsequent calls to recv for the
-        # message parts after the first don't block
-        with self._eventlet_recv_lock:
-            return _Socket_recv_pyobj(self, flags, copy, track)
+            return _Socket_recv_string(self, flags, encoding)
 
     @_wraps(_Socket.recv_json)
-    def recv_json(self, flags=0, copy=True, track=False):
+    def recv_json(self, flags=0, **kwargs):
         """A recv_json method that's safe to use when multiple
         greenthreads are calling send, send_json, recv and
         recv_json on the same socket.
         """
         if flags & NOBLOCK:
-            return _Socket_recv_json(self, flags, copy, track)
+            return _Socket_recv_json(self, flags, **kwargs)
 
         # acquire lock here so the subsequent calls to recv for the
         # message parts after the first don't block
         with self._eventlet_recv_lock:
-            return _Socket_recv_json(self, flags, copy, track)
+            return _Socket_recv_json(self, flags, **kwargs)
+
+    @_wraps(_Socket.recv_pyobj)
+    def recv_pyobj(self, flags=0):
+        """A recv_pyobj method that's safe to use when multiple
+        greenthreads are calling send, send_pyobj, recv and
+        recv_pyobj on the same socket.
+        """
+        if flags & NOBLOCK:
+            return _Socket_recv_pyobj(self, flags)
+
+        # acquire lock here so the subsequent calls to recv for the
+        # message parts after the first don't block
+        with self._eventlet_recv_lock:
+            return _Socket_recv_pyobj(self, flags)
